# Copyright 2014 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log

from neutron.common import topics
from neutron.extensions import portbindings
from neutron.plugins.common import constants as p_constants
from neutron.plugins.ml2 import driver_api as api
from neutron.plugins.ml2.drivers import mech_agent
from neutron.plugins.ml2 import rpc

from networking_powervm.plugins.ibm.agent.powervm import constants as pconst

LOG = log.getLogger(__name__)


class PvmSEAMechanismDriver(mech_agent.SimpleAgentMechanismDriverBase):
    """Attach to networks using PowerVM Shared Ethernet agent.

    The PvmSEAMechanismDriver integrates the ml2 plugin with the
    PowerVM Shared Ethernet Agent.
    """

    def __init__(self):
        super(PvmSEAMechanismDriver, self).__init__(
            pconst.AGENT_TYPE_PVM_SEA,
            pconst.VIF_TYPE_PVM_SEA,
            {portbindings.CAP_PORT_FILTER: False})
        self.rpc_publisher = rpc.AgentNotifierApi(topics.AGENT)

    def check_segment_for_agent(self, segment, agent):
        # TODO(thorst) Remove this in OpenStack Newton.  It can be assumed that
        # all agents are properly returning the mappings at that time.  The
        # agents started reporting the mappings in Mitaka.
        if self.get_mappings(agent):
            return (super(PvmSEAMechanismDriver, self).
                    check_segment_for_agent(segment, agent))
        else:
            LOG.debug("Checking segment: %s", segment)
            network_type = segment[api.NETWORK_TYPE]
            return network_type in ['vlan']

    def try_to_bind_segment_for_agent(self, context, segment, agent):
        # When this method is called, the parent should ideally be calling
        # down to the agent to state that the port was updated.  However,
        # it appears this isn't flowing properly.  This makes sure the
        # port is passed down to the agent.
        bindable = (super(PvmSEAMechanismDriver, self).
                    try_to_bind_segment_for_agent(context, segment, agent))
        if bindable:
            self.rpc_publisher.port_update(context._plugin_context,
                                           context._port,
                                           segment[api.NETWORK_TYPE],
                                           segment[api.SEGMENTATION_ID],
                                           segment[api.PHYSICAL_NETWORK])
        return bindable

    def get_allowed_network_types(self, agent=None):
        return [p_constants.TYPE_VLAN]

    def get_mappings(self, agent):
        return agent['configurations'].get('bridge_mappings', {})



# Copyright 2014, 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from pypowervm import const as pvm_const
from pypowervm import exceptions as pvm_exc
from pypowervm.helpers import log_helper as pvm_log
from pypowervm import util as pvm_util
from pypowervm.utils import retry as pvm_retry
from pypowervm.wrappers import logical_partition as pvm_lpar
from pypowervm.wrappers import managed_system as pvm_ms
from pypowervm.wrappers import network as pvm_net
from pypowervm.wrappers import virtual_io_server as pvm_vios

from networking_powervm._i18n import _LW
from networking_powervm.plugins.ibm.agent.powervm import exceptions as np_exc

LOG = logging.getLogger(__name__)

"""Provides a set of utilities for API interaction and Neutron."""


def get_host_uuid(adapter):
    """Get the System wrapper and its UUID for the (single) host.

    :param adapter: The pypowervm adapter.
    """
    syswraps = pvm_ms.System.wrap(adapter.read(pvm_ms.System.schema_type))
    if len(syswraps) != 1:
        raise np_exc.MultipleHostsFound(host_count=len(syswraps))
    return syswraps[0].uuid


def parse_sea_mappings(adapter, host_uuid, mapping):
    """This method will parse the sea mappings, and return a UUID map.

    The UUID of the NetworkBridges are required for modification of the
    VLANs that are bridged through the system (via the
    SharedEthernetAdapters). However, UUIDs are not user consumable.  This
    method will read in the string from the CONF file and return a mapping
    for the physical networks.

    Input:
     - <ph_network>:<sea>:<vios_name>,<next ph_network>:<sea2>:<vios_name>
     - Example: default:ent5:vios_lpar,speedy:ent6:vios_lpar

    Output:
    {
      'default': <Network Bridge UUID>, 'speedy': <Network Bridge 2 UUID>
    }

    :param adapter: The pypowervm adapter.
    :param host_uuid: The UUID for the host system.
    :param mapping: The mapping string as defined above to parse.
    :return: The output dictionary described above.
    """
    # Read all the network bridges.
    nb_wraps = list_bridges(adapter, host_uuid)

    if len(nb_wraps) == 0:
        raise np_exc.NoNetworkBridges()
    # Did the user specify the mapping?
    if mapping == '':
        return _parse_empty_bridge_mapping(nb_wraps)

    # Need to find a list of all the VIOSes names to hrefs
    vio_feed = adapter.read(pvm_ms.System.schema_type,
                            root_id=host_uuid,
                            child_type=pvm_vios.VIOS.schema_type,
                            xag=[pvm_const.XAG.VIO_NET])
    vio_wraps = pvm_vios.VIOS.wrap(vio_feed)

    # Response dictionary
    resp = {}

    # Parse the strings
    trios = mapping.split(',')
    for trio in trios:
        # Keys
        # 0 - physical network
        # 1 - SEA name
        # 2 - VIO name
        keys = trio.split(':')

        # Find the VIOS wrapper for the name
        vio_w = next(v for v in vio_wraps if v.name == keys[2])

        # For each network bridge, see if it maps to the SEA name/VIOS href
        matching_nb = None
        for nb_wrap in nb_wraps:
            for sea in nb_wrap.seas:
                if (sea.dev_name == keys[1] and
                        sea.vio_uri == vio_w.related_href):
                    # Found the matching SEA.
                    matching_nb = nb_wrap
                    break

        # Assuming we found a matching SEA, add it to the dictionary
        if matching_nb is not None:
            resp[keys[0]] = matching_nb.uuid
        else:
            raise np_exc.DeviceNotFound(dev=keys[1], vios=keys[2],
                                        phys_net=keys[0])

    return resp


def _parse_empty_bridge_mapping(bridges):
    """Will attempt to derive a bridge mapping if not specified.

    This method is invoked if there is no bridge mapping specified.
    If this happens, it will determine if there is a single Network Bridge
    on the system.  If so, it will assert that the default Neutron
    physical network resides on the singular Network Bridge.

    If there are multiple Network Bridges, an exception is raised.

    This does allow systems to not require the bridge mappings, but it
    is not ideal.

    :param bridges: A list of the network bridges returned via the API.
    :return: The bridge mapping, with a single physical network (default).
    :raises MultiBridgeNoMapping: Thrown if there are multiple Network
                                  Bridges on the system.
    """
    if len(bridges) > 1:
        raise np_exc.MultiBridgeNoMapping()

    LOG.warning(_LW('The bridge_mappings for the agent was not specified.  '
                    'There was exactly one Network Bridge on the system.  '
                    'Agent is assuming the default network is backed by the '
                    'single Network Bridge.'))
    return {'default': bridges[0].uuid}


def norm_mac(mac):
    """
    Will return a MAC Address that normalizes from the pypowervm format
    to the neutron format.

    That means that the format will be converted to lower case and will
    have colons added.

    :param mac: A pypowervm mac address.  Ex. 1234567890AB
    :returns: A mac that matches the standard neutron format.
              Ex. 12:34:56:78:90:ab
    """
    mac = mac.lower().replace(':', '')
    return ':'.join(mac[i:i + 2] for i in range(0, len(mac), 2))


def find_cna_for_mac(mac, client_adpts):
    """Returns the appropriate client adapter for a given mac address.

    :param mac: The mac address of the client adapter.
    :param client_adpts: The Client Adapters from pypowervm.
    :returns: The Client Adapter for the mac.  If one isn't found, then
              None will be returned.
    """
    mac = pvm_util.sanitize_mac_for_api(mac)

    for client_adpt in client_adpts:
        if client_adpt.mac == mac:
            return client_adpt

    # None was found.
    return None


def find_nb_for_cna(nb_wraps, client_adpt, vswitch_map):
    """
    Determines the NetworkBridge (if any) that is supporting a client
    adapter.

    :param nb_wraps: The network bridge wrappers on the system.
    :param client_adpt: The client adapter wrapper.
    :param vswitch_map: Maps the vSwitch IDs to URIs.
                        See 'get_vswitch_map'
    :return The Network Bridge wrapper that is hosting the client adapter.
            If there is not one, None is returned.
    """
    for nb_wrap in nb_wraps:
        # If the vSwitch ID doesn't match the vSwitch on the CNA...don't
        # process
        if vswitch_map.get(nb_wrap.vswitch_id) != client_adpt.vswitch_uri:
            continue

        # If the VLAN is not on the network bridge, then do not process.
        if not nb_wrap.supports_vlan(client_adpt.pvid):
            continue

        # At this point, the client adapter is supported by this network
        # bridge
        return nb_wrap

    # No valid network bridge
    return None


@pvm_retry.retry()
def get_vswitch_map(adapter, host_uuid):
    """Returns a dictionary of vSwitch IDs to their URIs.

    Ex. {'0': 'https://.../VirtualSwitch/<UUID>'}

    :param adapter: The pypowervm adapter.
    :param host_uuid: The UUID for the host system.
    """
    vsw_feed = adapter.read(pvm_ms.System.schema_type, root_id=host_uuid,
                            child_type=pvm_net.VSwitch.schema_type)
    vswitches = pvm_net.VSwitch.wrap(vsw_feed)
    resp = {}
    for vswitch in vswitches:
        resp[vswitch.switch_id] = vswitch.related_href
    return resp


def list_lpar_uuids(adapter, host_uuid):
    """Returns a list of all of the VM UUIDs.

    :param adapter: The pypowervm adapter.
    :param host_uuid: The UUID for the host system.
    :return: List of UUIDs for the VMs.
    """
    return [x.uuid for x in _list_vm_entries(adapter, host_uuid)]


def list_cnas(adapter, host_uuid, lpar_uuid=None):
    """Lists all of the Client Network Adapters for the running VMs.

    :param adapter: The pypowervm adapter.
    :param host_uuid: The UUID for the host system.
    :param lpar_uuid: (Optional) If specified, will only return the CNA's for
                      a given LPAR ID.
    """
    # Get the UUIDs of the VMs to query for.
    if lpar_uuid:
        vm_uuids = [lpar_uuid]
    else:
        vm_uuids = [x.uuid for x in _list_vm_entries(adapter, host_uuid)]

    # Loop through the VMs
    total_cnas = []
    for vm_uuid in vm_uuids:
        total_cnas.extend(_find_cnas(adapter, vm_uuid))

    return total_cnas


def _remove_log_helper(adapter):
    # Remove the log handler from the adapter so we don't log missing VMs
    # Pulling the helpers makes a copy
    helpers = adapter.helpers
    try:
        helpers.remove(pvm_log.log_helper)
    except ValueError:
        # It's not an error if we didn't find it since we don't want it.
        pass
    return helpers


@pvm_retry.retry()
def _find_cnas(adapter, vm_uuid):
    try:
        # Extend the array to include the response
        vm_cna_feed_resp = adapter.read(
            pvm_lpar.LPAR.schema_type, root_id=vm_uuid,
            child_type=pvm_net.CNA.schema_type,
            helpers=_remove_log_helper(adapter))
        return pvm_net.CNA.wrap(vm_cna_feed_resp)
    except pvm_exc.HttpError as e:
        # If it is a 404 (not found) then just skip.
        if e.response is not None and e.response.status == 404:
            return []
        else:
            raise


@pvm_retry.retry()
def _list_vm_entries(adapter, host_uuid):
    """
    Returns a List of all of the Client (non-VIOS) VMs on the system.
    Does not take into account whether or not it is managed by
    OpenStack.

    :param adapter: The pypowervm adapter.
    :param host_uuid: The UUID for the host system.
    """
    vm_feed = adapter.read(pvm_ms.System.schema_type, root_id=host_uuid,
                           child_type=pvm_lpar.LPAR.schema_type)
    vm_entries = vm_feed.feed.entries
    vms = []
    for vm_entry in vm_entries:
        vms.append(pvm_lpar.LPAR.wrap(vm_entry))
    return vms


@pvm_retry.retry()
def list_bridges(adapter, host_uuid):
    """
    Queries for the NetworkBridges on the system.  Will return the
    wrapper objects that describe Network Bridges.

    :param adapter: The pypowervm adapter.
    :param host_uuid: The UUID for the host system.
    """
    resp = adapter.read(pvm_ms.System.schema_type, root_id=host_uuid,
                        child_type=pvm_net.NetBridge.schema_type)
    net_bridges = pvm_net.NetBridge.wrap(resp)

    if len(net_bridges) == 0:
        LOG.warning(_LW('No NetworkBridges detected on the host.'))

    return net_bridges


def update_cna_pvid(cna, pvid):
    """This method will update the CNA with a new PVID.

    Will handle the retry logic surrounding this.  As the CNA may have
    come from old data.

    :param cna: The CNA wrapper (client network adapter).
    :param pvid: The new pvid to put on the wrapper.
    """

    def _cna_argmod(this_try, max_tries, *args, **kwargs):
        # Refresh the CNA to get a new etag
        LOG.debug("Attempting to re-query a CNA to get latest etag.")
        cna = args[0]
        cna.refresh()
        return args, kwargs

    @pvm_retry.retry(argmod_func=_cna_argmod)
    def _func(cna, pvid):
        cna.pvid = pvid
        cna.update()

    # Run the function (w/ retry) to update the PVID
    _func(cna, pvid)



# Copyright 2014, 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

import mock

from networking_powervm.plugins.ibm.agent.powervm import agent_base
from networking_powervm.plugins.ibm.agent.powervm import sea_agent
from networking_powervm.tests.unit.plugins.ibm.powervm import base
from pypowervm.tests import test_fixtures as pvm_fx

from neutron.common import constants as q_const
from neutron import context as ctx


def FakeClientAdpt(mac, pvid, tagged_vlans):
    return mock.MagicMock(mac=mac, pvid=pvid, tagged_vlans=tagged_vlans)


def FakeNPort(mac, segment_id, phys_network):
    device = {'physical_network': phys_network, 'segmentation_id': segment_id}
    return mock.Mock(mac=mac, segmentation_id=segment_id, rpc_device=device,
                     physical_network=phys_network, lpar_uuid='lpar_uuid')


def FakeNB(uuid, pvid, tagged_vlans, addl_vlans):
    m = mock.MagicMock()
    m.uuid = uuid

    lg = mock.MagicMock()
    lg.pvid = pvid
    lg.tagged_vlans = tagged_vlans

    vlans = [pvid]
    vlans.extend(tagged_vlans)
    vlans.extend(addl_vlans)
    m.list_vlans.return_value = vlans

    m.load_grps = [lg]
    return m


class FakeException(Exception):
    """Used to indicate an error in an API the agent calls."""
    pass


class SEAAgentTest(base.BasePVMTestCase):

    def setUp(self):
        super(SEAAgentTest, self).setUp()

        self.adpt = self.useFixture(
            pvm_fx.AdapterFx(traits=pvm_fx.LocalPVMTraits)).adpt

        with mock.patch('networking_powervm.plugins.ibm.agent.powervm.utils.'
                        'get_host_uuid'),\
                mock.patch('networking_powervm.plugins.ibm.agent.'
                           'powervm.utils.parse_sea_mappings') as mappings:
            mappings.return_value = {'default': 'nb_uuid'}
            self.agent = sea_agent.SharedEthernetNeutronAgent()
            self.agent.adapter = self.adpt

    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.utils.'
                'parse_sea_mappings')
    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.utils.'
                'get_host_uuid')
    def test_init(self, mock_get_host_uuid, mock_parse_mapping):
        """Verifies the integrity of the agent after being initialized."""
        mock_get_host_uuid.return_value = 'host_uuid'
        temp_agent = sea_agent.SharedEthernetNeutronAgent()
        self.assertEqual('networking-powervm-sharedethernet-agent',
                         temp_agent.agent_state.get('binary'))
        self.assertEqual(q_const.L2_AGENT_TOPIC,
                         temp_agent.agent_state.get('topic'))
        self.assertEqual(True, temp_agent.agent_state.get('start_flag'))
        self.assertEqual('PowerVM Shared Ethernet agent',
                         temp_agent.agent_state.get('agent_type'))

    def test_updated_ports(self):
        """
        Validates that the updated ports list can be added to and reset
        properly as needed.
        """
        self.assertEqual(0, len(self.agent._list_updated_ports()))

        self.agent._update_port({'mac_address': 'aa'})
        self.agent._update_port({'mac_address': 'bb'})

        self.assertEqual(2, len(self.agent._list_updated_ports()))

        # This should now be reset back to zero length
        self.assertEqual(0, len(self.agent._list_updated_ports()))

    def test_report_state(self):
        """"Validates that the report state functions properly."""
        # Make sure we had a start flag before the first report
        self.assertIsNotNone(self.agent.agent_state.get('start_flag'))

        # Mock up the state_rpc
        self.agent.state_rpc = mock.Mock()
        self.agent.context = mock.Mock()

        # run the code
        self.agent._report_state()

        # Devices are not set
        configs = self.agent.agent_state.get('configurations')
        self.assertEqual(0, configs['devices'])

        # Make sure we flipped to None after the report.  Also
        # indicates that we hit the last part of the method and didn't
        # fail.
        self.assertIsNone(self.agent.agent_state.get('start_flag'))

    @mock.patch('pypowervm.tasks.network_bridger.ensure_vlans_on_nb')
    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.utils')
    def test_provision_devices(self, mock_utils, mock_ensure):
        """Validates that the provision is invoked with batched VLANs."""
        self.agent.api_utils = mock_utils
        self.agent.br_map = {'default': 'nb_uuid'}
        self.agent.pvid_updater = mock.MagicMock()

        # Invoke
        self.agent.provision_devices([FakeNPort('aa', 20, 'default'),
                                      FakeNPort('bb', 22, 'default')])

        # Validate that both VLANs are in one call
        mock_ensure.assert_called_once_with(mock.ANY, mock.ANY, 'nb_uuid',
                                            {20, 22})

        # Validate that the PVID updates were completed
        self.assertEqual(2, self.agent.pvid_updater.add.call_count)

    @mock.patch('pypowervm.tasks.network_bridger.ensure_vlans_on_nb')
    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.utils')
    def test_provision_devices_fails(self, mock_utils, mock_ensure):
        """Validates that behavior of a failed VLAN provision."""
        self.agent.api_utils = mock_utils
        self.agent.br_map = {'default': 'nb_uuid'}
        self.agent.pvid_updater = mock.MagicMock()

        # Have the ensure throw some exception
        mock_ensure.side_effect = FakeException()

        # Invoke
        self.assertRaises(FakeException, self.agent.provision_devices,
                          [FakeNPort('aa', 20, 'default'),
                           FakeNPort('bb', 22, 'default')])

        # Validate that both VLANs are in one call.  Should still occur even
        # though no exception.
        mock_ensure.assert_called_once_with(mock.ANY, mock.ANY, 'nb_uuid',
                                            {20, 22})

        # However, the pvid updater should not be invoked.
        self.assertEqual(0, self.agent.pvid_updater.add.call_count)

    @mock.patch('pypowervm.tasks.network_bridger.remove_vlan_from_nb')
    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.sea_agent.'
                'SharedEthernetNeutronAgent._get_nb_and_vlan')
    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.sea_agent.'
                'SharedEthernetNeutronAgent.provision_devices')
    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.utils.'
                'get_vswitch_map')
    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.utils.list_cnas')
    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.utils.'
                'list_bridges')
    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.utils.'
                'find_nb_for_cna')
    def test_heal_and_optimize(
            self, mock_find_nb_for_cna, mock_list_bridges, mock_list_cnas,
            mock_vs_map, mock_prov_devs, mock_get_nb_and_vlan,
            mock_nbr_remove):
        """Validates the heal and optimization code."""
        # Fake adapters already on system.
        adpts = [FakeClientAdpt('00', 30, []),
                 FakeClientAdpt('11', 31, [32, 33, 34])]
        mock_list_cnas.return_value = adpts

        # The neutron data.  These will be 'ensured' on the bridge.
        self.agent.plugin_rpc = mock.MagicMock()
        self.agent.plugin_rpc.get_devices_details_list.return_value = [
            FakeNPort('00', 20, 'default'), FakeNPort('22', 22, 'default')]

        self.agent.br_map = {'default': 'nb_uuid'}

        # State that there is a pending VLAN (47) that has yet to be applied
        self.agent.pvid_updater = mock.MagicMock()
        self.agent.pvid_updater.pending_vlans = {47}

        # Mock a provision request
        p_req = mock.Mock()
        mock_get_nb_and_vlan.return_value = ('nb2_uuid', 23)

        # Mock up network bridges.  VLANs 44, 45, and 46 should be deleted
        # as they are not required by anything.  VLAN 47 should be needed
        # as it is in the pending list.
        mock_nb1 = FakeNB('nb_uuid', 20, [], [])
        mock_nb2 = FakeNB('nb2_uuid', 40, [41, 42, 43], [44, 45, 46, 47])
        mock_list_bridges.return_value = [mock_nb1, mock_nb2]
        mock_find_nb_for_cna.return_value = mock_nb2

        # Invoke
        self.agent.heal_and_optimize(False, [p_req], [], [])

        # Verify.  One ensure call per net bridge.
        self.assertEqual(3, mock_nbr_remove.call_count)
        mock_prov_devs.assert_called_with([p_req])

    @mock.patch('pypowervm.tasks.network_bridger.remove_vlan_from_nb')
    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.sea_agent.'
                'SharedEthernetNeutronAgent.provision_devices')
    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.utils.'
                'get_vswitch_map')
    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.utils.list_cnas')
    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.utils.'
                'list_bridges')
    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.utils.'
                'find_nb_for_cna')
    def test_heal_and_optimize_no_remove(
            self, mock_find_nb_for_cna, mock_list_bridges, mock_list_cnas,
            mock_vs_map, mock_prov_devs, mock_nbr_remove):
        """Validates the heal and optimization code. No remove."""
        # Fake adapters already on system.
        adpts = [FakeClientAdpt('00', 30, []),
                 FakeClientAdpt('11', 31, [32, 33, 34])]
        mock_list_cnas.return_value = adpts

        # The neutron data.  These will be 'ensured' on the bridge.
        self.agent.plugin_rpc = mock.MagicMock()
        self.agent.plugin_rpc.get_devices_details_list.return_value = [
            FakeNPort('00', 20, 'default'), FakeNPort('22', 22, 'default')]

        self.agent.br_map = {'default': 'nb_uuid'}

        # State that there is a pending VLAN (47) that has yet to be applied
        self.agent.pvid_updater = mock.MagicMock()
        self.agent.pvid_updater.pending_vlans = {47}

        # Mock up network bridges.  VLANs 44, 45, and 46 should be deleted
        # as they are not required by anything.  VLAN 47 should be needed
        # as it is in the pending list.
        mock_nb1 = FakeNB('nb_uuid', 20, [], [])
        mock_nb2 = FakeNB('nb2_uuid', 40, [41, 42, 43], [44, 45, 46, 47])
        mock_list_bridges.return_value = [mock_nb1, mock_nb2]
        mock_find_nb_for_cna.return_value = mock_nb2

        # Set that we can't do the clean up
        cfg.CONF.set_override('automated_powervm_vlan_cleanup', False, 'AGENT')

        # Invoke
        self.agent.heal_and_optimize(False, [], [], [])

        # Verify.  One ensure call per net bridge.  Zero for the remove as that
        # has been flagged to not clean up.
        self.assertEqual(0, mock_nbr_remove.call_count)
        self.assertEqual(1, mock_prov_devs.call_count)

    @mock.patch('oslo_service.loopingcall.FixedIntervalLoopingCall')
    @mock.patch.object(ctx, 'get_admin_context_without_session',
                       return_value=mock.Mock())
    def test_setup_rpc(self, admin_ctxi, mock_loopingcall):
        """Validates that the setup_rpc method is properly invoked."""
        cfg.CONF.AGENT = mock.Mock()
        cfg.CONF.AGENT.report_interval = 5

        # Derives the instance that will be returned when a new loopingcall
        # is made.  Used for verification
        instance = mock_loopingcall.return_value

        # Run the method to completion
        self.agent.setup_rpc()

        # Make sure that the loopingcall had an interval of 5.
        instance.start.assert_called_with(interval=5)

    def test_get_nb_and_vlan(self):
        """Be sure nb uuid and vlan parsed from dev properly."""
        dev = FakeNPort('a', 100, 'physnet1')
        self.agent.br_map = {'physnet1': 'uuid1'}
        uuid, vlan = self.agent._get_nb_and_vlan(dev.rpc_device)
        self.assertEqual('uuid1', uuid)
        self.assertEqual(100, vlan)


class PVIDLooperTest(base.BasePVMTestCase):

    def setUp(self):
        super(PVIDLooperTest, self).setUp()

        self.mock_agent = mock.MagicMock()
        self.looper = sea_agent.PVIDLooper(self.mock_agent)

    def build_update_req(self, mac, lpar, vlan):
        dev = {'mac_address': mac, 'physical_network': 'default',
               'segmentation_id': vlan, 'device_owner': 'nova:compute'}
        preq = agent_base.ProvisionRequest(dev, lpar)
        return sea_agent.UpdateVLANRequest(preq)

    def test_add(self):
        req = sea_agent.UpdateVLANRequest(mock.MagicMock())
        self.assertEqual(0, len(self.looper.requests))
        self.looper.add(req)
        self.assertEqual(1, len(self.looper.requests))

    def test_add_with_duplicate(self):
        # Add a baseline request
        self.looper.add(self.build_update_req('aa', '1', 1))
        self.assertEqual(1, len(self.looper.requests))

        # Add a second request.  Same LPAR, different mac and vlan
        self.looper.add(self.build_update_req('bb', '1', 2))
        self.assertEqual(2, len(self.looper.requests))

        # Add a third request.  Different LPAR and VLAN, same mac. (VERY RARE,
        # maybe not possible in Neutron).
        self.looper.add(self.build_update_req('aa', '2', 1))
        self.assertEqual(3, len(self.looper.requests))

        # Add some duplicates.  Same mac and lpar.
        self.looper.add(self.build_update_req('aa', '1', 1))
        self.looper.add(self.build_update_req('bb', '1', 2))
        self.assertEqual(3, len(self.looper.requests))

    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.utils.'
                'list_lpar_uuids')
    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.utils.'
                'list_cnas')
    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.utils.'
                'update_cna_pvid')
    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.utils.'
                'find_cna_for_mac')
    def test_update(self, mock_find_cna_for_mac, mock_update_cna_pvid,
                    mock_list_cnas, mock_uuids):
        req = sea_agent.UpdateVLANRequest(FakeNPort('a', 27, 'phys_net'))
        self.looper.add(req)

        # Mock the element returned
        mock_cna = mock.MagicMock()
        mock_find_cna_for_mac.return_value = mock_cna
        mock_list_cnas.return_value = [mock_cna]
        mock_uuids.return_value = ['lpar_uuid']

        # Call the update
        self.looper.update()

        # Check to make sure the request was pulled off
        self.assertEqual(0, len(self.looper.requests))

        # Make sure the mock CNA had update called, and the vid set correctly
        mock_update_cna_pvid.assert_called_with(mock_cna, 27)

        # Make sure the port was updated
        self.assertFalse(self.mock_agent.update_device_down.called)
        self.assertTrue(self.mock_agent.update_device_up.called)

    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.utils.'
                'list_lpar_uuids')
    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.utils.'
                'list_cnas')
    def test_update_no_lpar(self, mock_list_cnas, mock_uuids):
        req = sea_agent.UpdateVLANRequest(FakeNPort('a', 27, 'phys_net'))
        self.looper.add(req)

        # Mock the element returned
        mock_uuids.return_value = []

        # Call the update
        self.looper.update()

        # Check to make sure the request was pulled off
        self.assertEqual(1, len(self.looper.requests))
        self.assertFalse(mock_list_cnas.called)

    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.utils.'
                'list_lpar_uuids')
    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.utils.'
                'list_cnas')
    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.utils.'
                'find_cna_for_mac')
    def test_update_err(self, mock_find_cna_for_mac, mock_list_cnas,
                        mock_list_lpar_uuids):
        """Tests that the loop will error out after multiple loops."""
        mock_net_dev = FakeNPort('a', 1000, 'phys_net')

        req = sea_agent.UpdateVLANRequest(mock_net_dev)
        self.looper.add(req)

        # Mock the element returned
        mock_list_lpar_uuids.return_value = 'lpar_uuid'
        mock_list_cnas.return_value = [mock.Mock(mac='AABBCCDDEEFF', pvid=5)]
        mock_find_cna_for_mac.return_value = None

        loop_count = cfg.CONF.AGENT.pvid_update_loops

        for i in range(1, loop_count + 1):
            # Call the update
            self.looper.update()

            # Check to make sure the request was pulled off
            required_count = 1 if i < loop_count else 0
            self.assertEqual(required_count, len(self.looper.requests))

        self.assertTrue(self.mock_agent.update_device_down.called)
        self.assertFalse(self.mock_agent.update_device_up.called)



# Copyright 2014 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from networking_powervm.plugins.ibm.agent.powervm import exceptions as np_exc
from networking_powervm.plugins.ibm.agent.powervm import utils
from networking_powervm.tests.unit.plugins.ibm.powervm import base

from pypowervm import const as pvm_const
from pypowervm import exceptions as pvm_exc


class UtilsTest(base.BasePVMTestCase):
    """Tests the utility functions for the Shared Ethernet Adapter Logic."""

    def __cna(self, mac):
        """Create a Client Network Adapter mock."""

        class FakeCNA(object):

            @property
            def slot(self):
                return 1

            @property
            def mac(self):
                return mac

            @property
            def pvid(self):
                return 1

        return FakeCNA()

    def test_find_cna_for_mac(self):
        cna1 = self.__cna("1234567890AB")
        cna2 = self.__cna("123456789012")

        self.assertEqual(cna1, utils.find_cna_for_mac("1234567890AB",
                                                      [cna1, cna2]))
        self.assertEqual(None, utils.find_cna_for_mac("9876543210AB",
                                                      [cna1, cna2]))

    def test_norm_mac(self):
        EXPECTED = "12:34:56:78:90:ab"
        self.assertEqual(EXPECTED, utils.norm_mac("12:34:56:78:90:ab"))
        self.assertEqual(EXPECTED, utils.norm_mac("1234567890ab"))
        self.assertEqual(EXPECTED, utils.norm_mac("12:34:56:78:90:AB"))
        self.assertEqual(EXPECTED, utils.norm_mac("1234567890AB"))

    @mock.patch('pypowervm.wrappers.network.NetBridge.wrap')
    def test_list_bridges(self, mock_wrap):
        """Test that we can load the bridges in properly."""
        mock_wrap.return_value = ['br1', 'br2']
        mock_adpt = mock.Mock()
        mock_adpt.read = mock.Mock()

        # Assert that two are read in
        bridges = utils.list_bridges(mock_adpt, 'host_uuid')
        self.assertEqual(2, len(bridges))

    @mock.patch('pypowervm.wrappers.logical_partition.LPAR.wrap')
    def test_list_vm_entries(self, mock_wrap):
        """Validates that VMs can be iterated on properly."""
        feed = mock.Mock(object)
        feed.entries = ['1', '2', '3']
        vm_feed = mock.Mock(object)
        vm_feed.feed = feed
        adpt = mock.Mock()
        adpt.read = mock.Mock(return_value=vm_feed)

        # Mock the pypowervm wrapper to just return what it's passed
        mock_wrap.side_effect = lambda arg: arg

        # List the VMs and make some assertions
        vm_list = utils._list_vm_entries(adpt, 'host_uuid')
        self.assertEqual(3, len(vm_list))
        for vm in vm_list:
            self.assertIsNotNone(vm)
        self.assertEqual(3, mock_wrap.call_count)

    @mock.patch('pypowervm.wrappers.network.VSwitch.wrap')
    def test_get_vswitch_map(self, mock_wrap):
        # Create mocks
        mock_adpt = mock.Mock()
        mock_adpt.read = mock.Mock()
        vswitch = mock.Mock()
        vswitch.related_href = 'http://9.1.2.3/test'
        vswitch.switch_id = 0
        mock_wrap.return_value = [vswitch]

        # Run test and verify results
        resp = utils.get_vswitch_map(mock_adpt, 'host_uuid')
        self.assertEqual('http://9.1.2.3/test', resp[0])
        mock_adpt.read.assert_called_once_with('ManagedSystem',
                                               child_type='VirtualSwitch',
                                               root_id='host_uuid')

    def test_find_nb_for_cna(self):
        # Create mocks
        nb_wrap = mock.Mock()
        nb_wrap.vswitch_id = '0'
        nb_wrap.supports_vlan = mock.Mock(return_value=True)
        nb_wraps = [nb_wrap]

        vswitch_map = {'0': 'http://9.1.2.3/vs1',
                       '1': 'http://9.1.2.3/vs2'}

        mock_client_adpt = mock.MagicMock()
        mock_client_adpt.vswitch_uri = ('http://9.1.2.3/vs1')

        # Should have a proper URI, so it should match
        resp = utils.find_nb_for_cna(nb_wraps, mock_client_adpt, vswitch_map)
        self.assertIsNotNone(resp)

        # Should not match if we change the vswitch URI
        mock_client_adpt.vswitch_uri = "Fake"
        resp = utils.find_nb_for_cna(nb_wraps, mock_client_adpt, vswitch_map)
        self.assertIsNone(resp)

    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.utils.'
                '_list_vm_entries')
    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.utils.'
                '_find_cnas')
    @mock.patch('pypowervm.wrappers.network.CNA.wrap')
    def test_list_cnas(self, mock_cna_wrap, mock_find_cnas, mock_list_vms):
        """Validates that the CNA's can be iterated against."""

        # Override the VM Entries with a fake CNA
        class FakeVM(object):
            @property
            def uuid(self):
                return 'fake_uuid'
        vm = FakeVM()

        def list_vms(adapter, host_uuid):
            return [vm]

        mock_find_cnas.return_value = [1]
        mock_list_vms.side_effect = list_vms
        mock_cna_wrap.return_value = ['mocked']

        # Get the CNAs and validate
        cnas = utils.list_cnas(None, 'host_uuid')
        self.assertEqual(1, len(cnas))

    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.utils.'
                'list_bridges')
    @mock.patch('pypowervm.wrappers.virtual_io_server.VIOS.wrap')
    def test_parse_sea_mappings(self, mock_wrap, mock_list_br):
        # Create mocks
        class FakeVIOS(object):
            @property
            def name(self):
                return '21-25D0A'

            @property
            def related_href(self):
                return 'https://9.1.2.3/vios1'
        mock_sea = mock.Mock()
        mock_sea.dev_name = 'ent8'
        mock_sea.vio_uri = 'https://9.1.2.3/vios1'
        nb_wrap = mock.Mock()
        nb_wrap.seas = [mock_sea]
        nb_wrap.uuid = '764f3423-04c5-3b96-95a3-4764065400bd'
        nb_wraps = [nb_wrap]
        mock_list_br.return_value = nb_wraps
        mock_adpt = mock.MagicMock()
        mock_wrap.return_value = [FakeVIOS()]

        # Run actual test
        resp = utils.parse_sea_mappings(mock_adpt, 'host_uuid',
                                        'default:ent8:21-25D0A')

        # Verify results
        self.assertEqual(1, len(resp.keys()))
        self.assertIn('default', resp)
        self.assertEqual('764f3423-04c5-3b96-95a3-4764065400bd',
                         resp['default'])

    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.utils.'
                'list_bridges')
    def test_parse_sea_mappings_no_bridges(self, mock_list_br):
        mock_list_br.return_value = []
        self.assertRaises(np_exc.NoNetworkBridges, utils.parse_sea_mappings,
                          None, 'host_uuid', '1:2:3')

    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.utils.'
                '_parse_empty_bridge_mapping')
    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.utils.'
                'list_bridges')
    def test_parse_call_to_empty_bridge(self, mock_list_br, mock_empty):
        mock_list_br.return_value = ['br1']

        utils.parse_sea_mappings(None, 'host_uuid', '')

        # Make sure the _parse_empty_bridge_mapping method was called
        self.assertEqual(1, mock_empty.call_count)

    def test_parse_empty_bridge_mappings(self):
        proper_wrap = mock.MagicMock()
        proper_wrap.uuid = '5'
        resp = utils._parse_empty_bridge_mapping([proper_wrap])

        self.assertEqual({'default': '5'}, resp)

        # Try the failure path
        self.assertRaises(np_exc.MultiBridgeNoMapping,
                          utils._parse_empty_bridge_mapping,
                          [proper_wrap, mock.Mock()])

    def test_update_cna_pvid(self):
        """Validates the update_cna_pvid method."""
        def build_mock():
            # Need to rebuild.  Since it returns itself a standard reset will
            # recurse infinitely.
            cna = mock.MagicMock()
            cna.refresh.return_value = cna
            return cna

        # Attempt happy path
        cna = build_mock()
        utils.update_cna_pvid(cna, 5)
        self.assertEqual(5, cna.pvid)
        self.assertEqual(1, cna.update.call_count)

        # Raise an error 3 times and make sure it eventually re-raises the root
        # etag exception
        cna = build_mock()
        err_resp = mock.MagicMock()
        err_resp.status = pvm_const.HTTPStatus.ETAG_MISMATCH
        error = pvm_exc.HttpError(err_resp)

        cna.update.side_effect = [error, error, error]
        self.assertRaises(pvm_exc.HttpError, utils.update_cna_pvid, cna, 5)
        self.assertEqual(3, cna.update.call_count)
        self.assertEqual(2, cna.refresh.call_count)

        # Raise an error 2 times and then eventually works
        cna = build_mock()
        cna.update.side_effect = [error, error, None]
        utils.update_cna_pvid(cna, 5)
        self.assertEqual(3, cna.update.call_count)
        self.assertEqual(2, cna.refresh.call_count)

        # Immediate re-raise of different type of exception
        cna = build_mock()
        err_resp.status = pvm_const.HTTPStatus.UNAUTHORIZED
        cna.update.side_effect = pvm_exc.HttpError(err_resp)

        self.assertRaises(pvm_exc.HttpError, utils.update_cna_pvid, cna, 5)
        self.assertEqual(1, cna.update.call_count)
        self.assertEqual(0, cna.refresh.call_count)
